<?php

namespace Drupal\media_thumbnails_svg\Plugin\MediaThumbnail;

use Drupal\Core\File\FileSystemInterface;
use Drupal\media_thumbnails\Plugin\MediaThumbnailBase;
use SVG\SVG;

/**
 * Media thumbnail plugin for svg documents.
 *
 * @\Drupal\media_thumbnails\Annotation\MediaThumbnail(
 *   id = "media_thumbnail_svg",
 *   label = @Translation("Media Thumbnail SVG"),
 *   mime = {
 *     "image/svg",
 *     "image/svg+xml",
 *   }
 * )
 */
class MediaThumbnailSVG extends MediaThumbnailBase {

  /**
   * The thumbnail width.
   *
   * @var int
   */
  protected $width;

  /**
   * The thumbnail background colour.
   *
   * @var string
   */
  protected $bgColor;

  /**
   * Creates a managed thumbnail file using the passed source file uri.
   *
   * {@inheritdoc}
   */
  public function createThumbnail($sourceUri) {

    // Not all rasterizers support stream wrappers, use absolute paths.
    $path = $this->fileSystem->realpath($sourceUri);

    // Get width and background color, if any.
    $this->bgColor = $this->configuration['bgcolor_active'] ? $this->configuration['bgcolor_value'] : 'transparent';
    $this->width = $this->configuration['width'] ?? 500;

    // Create a thumbnail image blob using the best rasterizer available.
    if ($this->binaryExists('gm') && strpos(shell_exec('gm'), 'GraphicsMagick') !== FALSE) {
      $image = $this->createThumbnailGraphicsMagick($path);
    }
    elseif ($this->binaryExists('convert') && strpos(shell_exec('convert'), 'ImageMagick') !== FALSE) {
      $image = $this->createThumbnailImageMagick($path);
    }
    else {
      $image = $this->createThumbnailGd($path);
    }

    // Return a new managed file object using the generated thumbnail.
    return $image ? file_save_data($image, $sourceUri . '.png', FileSystemInterface::EXISTS_REPLACE) : NULL;
  }

  /**
   * Create thumbnail with GraphicsMagick.
   *
   * @param string $path
   *   The SVG image path.
   *
   * @return false|string
   *   The SVG as a PNG.
   */
  protected function createThumbnailGraphicsMagick($path) {

    $source = escapeshellarg($path);
    $target = escapeshellarg($path . '.png');

    $width = escapeshellarg($this->width);
    $bg_colour = escapeshellarg($this->bgColor);

    shell_exec(
      sprintf(
        'gm convert -background %s -size %s -quality 100 -strip %s %s',
        $bg_colour, $width, $source, $target
      )
    );
    $data = file_get_contents($path . '.png');
    if (!$data) {
      $this->logger->warning($this->t('Could not create png from svg using GM.'));
    }
    return $data;
  }

  /**
   * Create thumbnail with ImageMagick.
   *
   * @param string $path
   *   The SVG image path.
   *
   * @return false|string
   *   The SVG as a PNG.
   */
  protected function createThumbnailImageMagick($path) {

    $source = escapeshellarg($path);
    $target = escapeshellarg($path . '.png');

    $width = escapeshellarg($this->width);
    $bg_colour = escapeshellarg($this->bgColor);

    shell_exec(
      sprintf(
        'convert -background %s -density %s -thumbnail %s -quality 100 -strip %s %s',
        $bg_colour, $width, $width, $source, $target
      )
    );
    $data = file_get_contents($path . '.png');
    if (!$data) {
      $this->logger->warning($this->t('Could not create png from svg using IM.'));
    }
    return $data;

  }

  /**
   * Create thumbnail with GD.
   *
   * @param string $path
   *   The SVG image path.
   *
   * @return false|string|null
   *   The SVG as a PNG.
   */
  protected function createThumbnailGd($path) {
    $image = SVG::fromFile($path);
    if (!$image) {
      $this->logger->warning($this->t('Media entity source file (svg) not found.'));
      return NULL;
    }

    // Create a raster image using the target width, keeping the aspect ratio.
    $width = $image->getDocument()->getWidth() ?: $image->getDocument()
      ->getViewBox()[2];
    $height = $image->getDocument()->getHeight() ?: $image->getDocument()
      ->getViewBox()[3];
    $ratio = $width && $height ? $height / $width : 1;
    $height = (int) ($this->width * $ratio);
    $raster_image = $image->toRasterImage($this->width, $height, $this->bgColor);

    // Create a new image thumbnail blob.
    ob_start();
    if (!imagepng($raster_image, NULL, 9)) {
      $this->logger->warning($this->t('Could not create png from svg using GD.'));
      ob_end_clean();
      return NULL;
    }
    return ob_get_clean();

  }

  /**
   * Returns whether a binary exists.
   *
   * @param string $binary
   *   The binary to check for.
   *
   * @return bool
   *   Whether the binary exists.
   */
  protected function binaryExists($binary) {
    if (defined('PHP_WINDOWS_VERSION_BUILD')) {
      $which = 'where';
      $dev_null = 'NUL';
    }
    else {
      // https://stackoverflow.com/a/677212
      $which = 'command -v';
      $dev_null = '/dev/null';
    }

    // Swallow up any stderr if the 'which' binary doesn't exist.
    $result = trim(
      shell_exec(
        sprintf(
          '%s %s 2> %s',
          escapeshellcmd($which),
          escapeshellarg($binary),
          escapeshellarg($dev_null)
        )
      )
    );

    return $result && is_executable($result);
  }

}
